import sys

sys.dont_write_bytecode = True

import multiprocessing
import shutil
import os

from pathlib import Path

from .compiler_functions import manage_pipes
from .compiler_functions import encrypt_strings, decrypt_strings
from .compiler_functions import preprocess
from .compiler_functions import better_range, replace_keywords, entry_point
from .compiler_functions import multi_line_lambda

from .qbstd import *

def main(args=None):
    if args is None:
        args = sys.argv[1:]

    if len(args) == 0:
        print("No file specified.")
        return

    file = args[0]

    if not file.endswith(".qb"):
        print("File must be a .qb file")
        return

    path = os.path.abspath(file)

    if not os.path.exists(path):
        print("File not found")
        return

    marked_for_deletion = []

    try:
        code = compile_to_qb(path)

        dirname = os.path.dirname(path)
        sys.path = [dirname] + sys.path

        for lib_path in Path(dirname).rglob("*.qb"):
            lib_path = str(lib_path)

            if lib_path == path:
                continue

            lib_code = compile_to_qb(lib_path)

            lib_path = lib_path.replace(".qb", ".py")
            write(lib_path, lib_code)

            marked_for_deletion.append(lib_path)

        # Get code ready to be executed
        sys.setrecursionlimit(2**31 - 1)
        sys.argv = [path] + args[1:]

        exec(code, globals())

        for file in marked_for_deletion:
            os.remove(file)

        pycache = os.path.join(dirname, "__pycache__")
        if os.path.exists(pycache):
            shutil.rmtree(pycache)

    except Exception as e:
        for file in marked_for_deletion:
            os.remove(file)

        dirname = os.path.dirname(path)

        pycache = os.path.join(dirname, "__pycache__")
        if os.path.exists(pycache):
            shutil.rmtree(pycache)

        raise e


def read(path):
    with open(path, "r") as file:
        return file.read()


def write(path, content):
    with open(path, "w") as file:
        file.write(content)


def compile_to_qb(path):
    code = read(path)

    encrypted = encrypt_strings(code)

    code = encrypted[0]
    encrypted_strings = encrypted[1]

    code = preprocess(code)
    code = multi_line_lambda(code)
    code = entry_point(code)
    code = replace_keywords(code)
    code = better_range(code)
    code = manage_pipes(code)

    code = decrypt_strings(code, encrypted_strings)

    return code


# if __name__ == "__main__":
#     main(sys.argv[1:])
#
#
# else:
#     print(
#         """⠁⡼⠋⠀⣆⠀⠀⣰⣿⣫⣾⢿⣿⣿⠍⢠⠠⠀⠀⢀⠰⢾⣺⣻⣿⣿⣿⣷⡀⠀
# ⣥⠀⠀⠀⠁⠀⠠⢻⢬⠁⣠⣾⠛⠁⠀⠀⠀⠀⠀⠀⠀⠐⠱⠏⡉⠙⣿⣿⡇⠀
# ⢳⠀⢰⡖⠀⠀⠈⠀⣺⢰⣿⢻⣾⣶⣿⣿⣶⣶⣤⣤⣴⣾⣿⣷⣼⡆⢸⣿⣧⠀
# ⠈⠀⠜⠈⣀⣔⣦⢨⣿⣿⣿⣾⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣅⣼⠛⢹⠀
# ⠀⠀⠀⠀⢋⡿⡿⣯⣭⡟⣟⣿⣿⣽⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⠀⡘⠀
# ⡀⠐⠀⠀⠀⣿⣯⡿⣿⣿⣿⣯⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⡿⠋⣉⢽⣿⡆⠀⠀
# ⢳⠀⠄⠀⢀⣿⣿⣿⣿⣿⣿⣿⠙⠉⠉⠉⠛⣻⢛⣿⠛⠃⠀⠐⠛⠻⣿⡇⠀⠀
# ⣾⠄⠀⠀⢸⣿⣿⡿⠟⠛⠁⢀⠀⢀⡄⣀⣠⣾⣿⣿⡠⣴⣎⣀⣠⣠⣿⡇⠀⠀
# ⣧⠀⣴⣄⣽⣿⣿⣿⣶⣶⣖⣶⣬⣾⣿⣾⣿⣿⣿⣿⣽⣿⣿⣿⣿⣿⣿⡇⠀⠀
# ⣿⣶⣈⡯⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⡿⠋⣹⢧⣿⣿⣿⣄⠙⢿⣿⣿⣿⠇⠀⠀
# ⠹⣿⣿⣧⢌⢽⣻⢿⣯⣿⣿⣿⣿⠟⣠⡘⠿⠟⠛⠛⠟⠛⣧⡈⠻⣾⣿⠀⠀⠀
# ⠀⠈⠉⣷⡿⣽⠶⡾⢿⣿⣿⣿⢃⣤⣿⣷⣤⣤⣄⣄⣠⣼⡿⢷⢀⣿⡏⠀⠀⠀
# ⠀⠀⢀⣿⣷⠌⣈⣏⣝⠽⡿⣷⣾⣏⣀⣉⣉⣀⣀⣀⣠⣠⣄⡸⣾⣿⠃⠀⠀⠀
# ⠀⣰⡿⣿⣧⡐⠄⠱⣿⣺⣽⢟⣿⣿⢿⣿⣍⠉⢀⣀⣐⣼⣯⡗⠟⡏⠀⠀⠀⠀
# ⣰⣿⠀⣿⣿⣴⡀⠂⠘⢹⣭⡂⡚⠿⢿⣿⣿⣿⡿⢿⢿⡿⠿⢁⣴⣿⣷⣶⣦⣤
#
#     Are you aight, mate?
# """
#     )



import ast


class PipeTransformer(ast.NodeTransformer):
    def visit_BinOp(self, node):
        if isinstance(node.op, ast.RShift):
            # Recursive call for the left node if it's also a BinOp
            left = self.visit(node.left) if isinstance(node.left, ast.BinOp) else node.left
            # Ensure that the right node is a function
            if isinstance(node.right, ast.Call):
                node.right.args.append(left)
                return node.right
            elif isinstance(node.right, ast.expr):  # Change here
                return ast.Call(func=node.right, args=[left], keywords=[])
            else:
                raise TypeError(get_pipe_error_text("right"))
        
        if isinstance(node.op, ast.LShift):
            # Recursive call for the right node if it's also a BinOp
            right = self.visit(node.right) if isinstance(node.right, ast.BinOp) else node.right
            # Ensure that the left node is a function
            if isinstance(node.left, ast.Call):
                node.left.args.append(right)
                return node.left
            elif isinstance(node.left, ast.expr):  # Change here
                return ast.Call(func=node.left, args=[right], keywords=[])
            else:
                raise TypeError(get_pipe_error_text("left"))
        
        return self.generic_visit(node)

def manage_pipes(code: str) -> str:
    tree = ast.parse(code)
    PipeTransformer().visit(tree)
    ast.fix_missing_locations(tree)

    return ast.unparse(tree)

def get_pipe_error_text(side: str) -> str:
    op = "|>" if side == "right" else "<|"
    error = f"The {side} operand of the '{op}' operator must be a function"
    return error


from dataclasses import dataclass
from warnings import warn


class Some[T]:
    def __init__(self, value: T):
        self.value = value

    def unwrap(self) -> T:
        return self.value


type maybe = Some | None
type either = Some | Exception


class collection():
    """A lazy evaluated collection, that behaves like a list, but immutable with some neat additional methods."""

    def __init__(self, *args):
        self.args = args

    def for_each(self, func):
        if len(self) == 0:
            return

        func(self[0])
        collection(*self[1:]).for_each(func)

    def map(self, func):
        if len(self) == 0:
            return self

        return collection(func(self[0]), *collection(*self[1:]).map(func))

    def filter(self, func):
        if len(self) == 0:
            return self

        if func(self[0]):
            return collection(self[0], *collection(*self[1:]).filter(func))
        else:
            return collection(*collection(*self[1:]).filter(func))
        
    def head(self):
        return self[0]
    
    def tail(self):
        return collection(*self[1:])

    @staticmethod
    def of(iterable):
        if hasattr(iterable, "__len__") and iterable.__len__() == float("inf"):
            warn("Passing an infinite range to collection.of will result in an infinite loop.")

        return collection(*iterable)

    def __iter__(self):
        return iter(self.args)

    def __getitem__(self, key):
        return self.args[key]

    def __len__(self):
        return len(self.args)

    def __repr__(self):
        return f"collection{self.args}"

    def __str__(self):
        return f"{self.args}"

    def __contains__(self, item):
        return item in self.args

    def __eq__(self, other):
        return self.args == other.args

    def __ne__(self, other):
        return self.args != other.args

    def __add__(self, other):
        return collection(*self.args, *other.args)


@dataclass()
class infinite_range:
    """Returns an object that can be used similar to a normal range object, but with no end."""

    start: int = 0
    step: int = 1

    def __iter__(self):
        return self

    def __next__(self):
        self.start += self.step
        return self.start - self.step

    def __getitem__(self, key):
        if isinstance(key, slice):
            start = key.start if key.start else self.start
            stop = key.stop if key.stop else float("inf")
            step = key.step if key.step else self.step

            return [self.start + i * self.step for i in range(start, stop, step)]
        else:
            return self.start + key * self.step

    def __len__(self):
        return float("inf")

    def __repr__(self):
        return f"infinite_range(start={self.start}, step={self.step})"

    def __str__(self):
        return f"range({self.start}, inf)"

    def __contains__(self, item):
        return (item - self.start) % self.step == 0



import re as regex

definition_regex = r"\b(fn)\s+(\w+\s*\()"
lambda_regex = r"fn\s*([^->]*)->"
up_one_regex = r"(\w+)\s?\+\+"
down_one_regex = r"(\w+)\s?\-\-"
not_equal_regex = r"<>"

range_function_regex_two_groups = regex.compile(r"(\w+)\.\.(\w+)")
range_function_regex_one_group = regex.compile(r"\.\.(\w+)")
range_function_regex_infinite = regex.compile(r"(\w+)\.\.")

entry_point_regex = regex.compile(r"@entrypoint(\(.*\))?\s+")

def replace_keywords(code):
    code = regex.sub(definition_regex, r"def \2", code)
    code = regex.sub(lambda_regex, r"lambda \1:", code)
    code = regex.sub(up_one_regex, r"\1 += 1", code)
    code = regex.sub(down_one_regex, r"\1 -= 1", code)
    code = regex.sub(not_equal_regex, r"!=", code)

    return code


def better_range(code):
    if ".." not in code:
        return code

    for b in regex.findall(range_function_regex_two_groups, code):
        code = code.replace(f"{b[0]}..{b[1]}", f"range({b[0]}, {b[1]})")

    for b in regex.findall(range_function_regex_one_group, code):
        code = code.replace(f"..{b}", f"range({b})")

    for b in regex.findall(range_function_regex_infinite, code):
        code = code.replace(f"{b}..", f"infinite_range({b})")

    return code


def entry_point(code):
    entry_points = list(regex.finditer(entry_point_regex, code))

    if len(entry_points) == 0:
        return code

    if len(entry_points) > 1:
        raise Exception("Multiple entry points found")

    entry_point_found = entry_points[0]
    args = entry_point_found.group(1)

    if args is None:
        args = "()"

    next_line = ""
    for c in code[entry_point_found.end():]:
        if c == "\n":
            break

        next_line += c

    signature = list(regex.finditer(definition_regex, next_line))

    if len(signature) == 0:
        raise Exception("No function signature found after entry point")

    function_name = signature[0].group(2)[:-1]

    # quick and dirty fix
    poetry___name__ = "qb.__main__"

    idiom = f"\nif __name__ == \"{poetry___name__}\":\n\t{function_name}{args}"

    code += idiom
    code = code.replace("@entrypoint", "")

    return code



def head[T](xs: T) -> T:
    return xs[0]


def tail[T](xs: T) -> T:
    return xs[1:]


def entrypoint() -> None:
    raise Exception("The entrypoint function is reserved as a placeholder for the compiler. Do not call it directly. Use the @entrypoint decorator to mark a function as the entry point.")


from warnings import warn


def preprocess(code):
    raise_errors_and_warnings(code)

    code = code.replace("|>", ">>")
    code = code.replace("<|", "<<")

    return code


def raise_errors_and_warnings(code):
    shifting_operators = ["<<", ">>"]
    shifting_error = "When bit shifting, use the 'lshift' and 'rshift' function instead of the '>>' and '<<' operators respectively"

    io_functions = ["print", "input"]
    io_warning = "When performing io-operations, use cout and cin instead of the 'print' and 'input' functions respectively"

    function_definitions = ["def", "lambda"]
    function_error = (
        "When defining a function, use the 'fn' keyword instead of 'def' or 'lambda' like so: 'fn x -> x ** 2'"
    )

    not_operator = "!="
    not_error = "When checking for inequality, use the '<>' operator instead of '!='"

    range_usage = "range"
    range_error = "When generating a range, use the '..' operator instead of the range function like so: '1..10'"

    for operator in shifting_operators:
        if operator in code:
            raise SyntaxError(shifting_error)

    for function in io_functions:
        if function in code:
            warn(SyntaxWarning(io_warning))

    for definition in function_definitions:
        if definition in code:
            warn(SyntaxWarning(function_error))

    if not_operator in code:
        warn(SyntaxWarning(not_error))

    if range_usage in code:
        warn(SyntaxWarning(range_error))


import re as regex
import time
import random


def encrypt_strings(code: str):
    """Encrypts strings in code"""

    # find all strings
    strings = regex.findall(r"[\"'].*[\"']", code)
    fstrings = regex.findall(r"f[\"'].*[\"']", code)

    encrypted_strings = {}

    for fstring in fstrings:
        # Split fstring into separate strings, excluding parts within curly braces
        # Needs fixing

        fstring_content = fstring[2:-1]
        parts = regex.split(r"{.*}", fstring_content)

        for part in parts:
            if part.strip() == "":
                continue

            part_id = f"part_{time.time_ns()}_{random.randint(0, 1000000)}"

            encrypted_strings[part_id] = part
            code = code.replace(part, part_id)

    for string in strings:
        string_content = string[1:-1]
        string_id = f"str_{time.time_ns()}_{random.randint(0, 1000000)}"

        encrypted_strings[string_id] = string_content
        code = code.replace(string_content, string_id)

    return code, encrypted_strings


def decrypt_strings(code: str, encrypted_strings: dict):
    """Decrypts strings in code"""

    # Reinsert curly braces
    fstring_ids = [id_ for id_ in encrypted_strings.keys() if id_.startswith("part_")]
    for fstring_id in fstring_ids:
        expressions = regex.findall(r"{(.*?)}", encrypted_strings[fstring_id])
        for expr in expressions:
            code = code.replace(
                fstring_id,
                fstring_id.replace(encrypted_strings[fstring_id], "{" + expr + "}"),
            )

    for string_id, original_string in encrypted_strings.items():
        code = code.replace(string_id, original_string)

    return code



import re as regex
import random
import time

def multi_line_lambda(code):
    multiline_lambda_regex = r"fn\s+(.*)->\s*{\|([\S\s]*)\|}"
    match = regex.search(multiline_lambda_regex, code)

    # Wenn keine Übereinstimmung gefunden wird, geben wir den Code zurück
    if match is None:
        return code

    _vars = match.group(1)
    _code = match.group(2)

    function_name = f"lambda_{time.time_ns()}_{random.randint(0, 1000000)}"

    body = "{|" + _code + "|}"

    definition = f"def {function_name}({_vars}):{_code}"
    call = f"{function_name}({_vars})"

    code = code.replace(body, call)

    index = match.start()
    code_to_search = code[:index][::-1]

    for c in code_to_search:
        if c == "\n":
            break

        index -= 1

    indent = ""
    for c in code[index:]:
        if c.strip() == "":
            indent += c

        else:
            break

    code = code[:index] + indent + definition + code[index + len(indent):]

    # Rufe die Funktion rekursiv auf, um weitere Übereinstimmungen zu finden
    return multi_line_lambda(code)